/*
 * Copyright (c) 2017, Arnon Yaari
 * All rights reserved.
 * Use of this source code is governed by a BSD-style license that can be
 * found in the LICENSE file.
 */

#ifndef __NET_CONNECTIONS_H__
#define __NET_CONNECTIONS_H__

#include <Python.h>

PyObject *psutil_net_connections(PyObject *self, PyObject *args);

#endif  // __NET_CONNECTIONS_H__
